/***********************************************************************
 * File: aggregate.c
 * 	(Extracted from
 *	   "$Id: aggregate.c,v 1.3 2009/05/12 13:46:09 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source file implements `aggregate()' API.
 *
 * Copyright (C) 2009 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "aggregate.h" /* definitions of `aggregate' API-s */

/* type definition of the result buffer slot */
typedef struct result_t {
	size_t	size;	/* size of the allocated result buffer */
	char	*bufp;	/* pointer to the allocated result buffer */
} Result_t;

/* definition of signs for `Number_t' */
typedef enum sign_e {
	  Sign_e_positive	/* sign is positive. */
	, Sign_e_negative	/* sign is negative. */
}	Sign_e;

/* type definition of the number for processing */
typedef struct number_t {
	const char	*ip;	/* pointer to integer part */
	const char	*dp;	/* pointer to decimal point */
	const char	*ep;	/* pointer to end of numeral string */
	Sign_e	sign;	/* sign of number */
} Number_t;

/* the Number_t structure `x' means zero or not? */
static int is_zero(Number_t	*x)
{
	int	result	= (((x->ip == x->dp) && x->ip == x->ep)? 1: 0);

	return	result;
}

/* make a Number_t structure as the numeral string */
/*@null@*/ static Number_t *make_number(			const char	*numeral_str
										, /*@out@*/	Number_t	*number
										)
{
	const char	*p = NULL;

	/* test: arguments are invalid? */
	if ((number == NULL) || numeral_str == NULL) {
		return	NULL; /* Error: invalid arguments */
	}

	/* initialize `number' structure */
	memset(number, (int)'\0', sizeof(*number));

	/* set pointers to `decimal point' and 'end of fraction part' */
	p	= numeral_str + strlen(numeral_str);
	if ((number->dp = strchr(numeral_str, '.')) == NULL) {
		while ((numeral_str < p) && *(p - 1) == ' ') {	--p;	}
		number->dp	= (char *)p;
	}
	while ((number->dp < p) && *(p - 1) == ' ') {	--p;	}
	while ((number->dp < p) && *(p - 1) == '0') {	--p;	}
	number->ep	= ((number->dp == (p - 1))? number->dp:	(char *)p);

	/* test: fraction part is filled with digits? */
	if (number->dp != number->ep) {
		for (p = number->dp + 1; p < number->ep; ++p) {
			if (!isdigit(*p)) {
				return	NULL; /* Error: invalid arguments */
	}	}	}

	/* set sign and pointer to `start of integer part' */
	p	= numeral_str;
	while ((p < number->dp) && *p == ' ') {	++p;	}
	number->sign	= (((p == number->dp) || *p != '-')?	Sign_e_positive
						:									Sign_e_negative
						);
	if ((*p == '+') || *p == '-') {	++p;	}
	if (p == number->dp) {
		number->ip	= (char *)p;
	}
	else {
		if ((*p != ' ') && !isdigit(*p)) {
			return	NULL; /* Error: invalid arguments */
		}
		while ((p < number->dp) && *p == ' ') {	++p;	}
		while ((p < number->dp) && *p == '0') {	++p;	}
		number->ip	= (char *)p;
	}

	/* test: integer part is filled with digits? */
	if (number->ip != number->dp) {
		for (p = number->ip; p < number->dp; ++p) {
			if (!isdigit(*p)) {
				return NULL; /* Error: invalid arguments */
	}	}	}

	/* test: numeral means zero? */
	if (is_zero(number) != 0) {
		/* zero has positive sign */
		number->sign	= Sign_e_positive;
	}

	return	number;	/* normal end */
}

/* compare two Number_t structures 'a' and 'b'
 * if `a' is less than `b', this returns -1.
 * if `a' equals `b', this returns zero.
 * if `a' is greater than `b', this returns 1.
 */
static int compare(Number_t	*a, Number_t	*b)
{
	size_t		a_len, b_len, limit;
	int			result;

	/* compare sign */
	if ((a->sign == Sign_e_negative) && b->sign == Sign_e_positive) {
			return	-1;	/* `a' is less than `b' */
	}
	else if ((b->sign == Sign_e_negative) && a->sign == Sign_e_positive) {
			return	1;	/* `a' is greater than `b' */
	}

	/* compare integer parts */
	a_len = (size_t)(a->dp - a->ip); b_len = (size_t)(b->dp - b->ip);
	if (a_len > b_len) {
		/* `a' is greater/less than `b' */
		return	((a->sign == Sign_e_positive)?	1:	-1);
	}
	else if (a_len < b_len) {
		/* `a' is less/greater than `b' */
		return	((a->sign == Sign_e_positive)?	-1:	1);
	}
	else if (a_len > 0) {
		if ((result = memcmp(a->ip, b->ip, a_len)) > 0) {
			/* `a' is greater/less than `b' */
			return	((a->sign == Sign_e_positive)?	1:	-1);
		}
		else if (result < 0) {
			/* `a' is less/greater than `b' */
			return	((a->sign == Sign_e_positive)?	-1:	1);
		}
	}

	/* compare fraction parts */
	a_len = (size_t)(a->ep - a->dp); b_len = (size_t)(b->ep - b->dp);
	limit	= ((a_len < b_len)? a_len: b_len);
	if (limit > 0) {
		if ((result = memcmp(a->dp , b->dp, limit)) > 0) {
			/* `a' is greater/less than `b' */
			return	((a->sign == Sign_e_positive)?	1:	-1);
		}
		else if (result < 0) {
			/* `a' is less/greater than `b' */
			return	((a->sign == Sign_e_positive)?	-1:	1);
		}
	}
	if (a_len > b_len) {
		/* `a' is greater/less than `b' */
		return	((a->sign == Sign_e_positive)?	1:	-1);
	}
	else if (a_len < b_len) {
		/* `a' is less/greater than `b' */
		return	((a->sign == Sign_e_positive)?	-1:	1);
	}

	return	0;	/* `a' equals `b' */
}

/* convert a digit code to integer */
static int digit2int(const char	*p)
{
	char	buf[2];

	buf[0] = *p; buf[1] = '\0'; return atoi(buf);
}

/* convert an integer to a digit code */
static char int2digit(int d)
{
	static char	digits[] =	{ '0', '1', '2', '3', '4'
							, '5', '6', '7', '8', '9'
							};
	return	digits[d];
}

/* add two Number_t structure `x' and `y'
 * signs of `x' and `y' must be `Sign_e_positive'
 */
/*@null@*/ static char *add(Number_t	*x
							, Number_t	*y
							, char		*result
							, size_t	result_size
							)
{
	const char	*xp, *yp;
	size_t		i, x_len, y_len, move_len;
	char		*rp;
	int			carry;

	/* set end of string */
	rp = result + (result_size - 1); *rp = '\0';

	/* process fraction parts */
	if ((x->ep != x->dp) || y->ep != y->dp) {
		if (x->ep == x->dp) {
			/* `x' has no fraction parts, `y' has it */
			if ((move_len = (size_t)(y->ep - y->dp)) > (result_size - 1)) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			rp	-= move_len;
			memcpy(rp, y->dp, move_len);
			xp	= x->dp; yp = y->dp; carry = 0;
		}
		else if (y->ep == y->dp) {
			/* `y' has no fraction part, `x' has it */
			if ((move_len = (size_t)(x->ep - x->dp)) > (result_size - 1)) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			rp	-= move_len;
			memcpy(rp, x->dp, move_len);
			xp	= x->dp; yp = y->dp; carry = 0;
		}
		else {
			/* adding the fraction digits */
			x_len = (size_t)(x->ep - x->dp)
			; y_len = (size_t)(y->ep - y->dp)
			;
			if (x_len > y_len) {
				/* `x' fraction part is longer than `y's */
				if ((move_len = x_len - y_len) > (result_size - 1)) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				rp	-= move_len;
				xp	= x->dp + (x_len - move_len);
				yp	= y->dp + y_len;
				memcpy(rp, xp, move_len);
				move_len	= y_len;
			}
			else if (x_len < y_len) {
				/* `y' fraction part is longer than `x's */
				if ((move_len = y_len - x_len) > (result_size - 1)) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				rp	-= move_len;
				xp	= x->dp + x_len;
				yp	= y->dp + (y_len - move_len);
				memcpy(rp, yp, move_len);
				move_len	= x_len;
			}
			else {
				/* 'x' fraction part is such length as `y's */
				xp	= x->dp + x_len;
				yp	= y->dp + y_len;
				move_len	= x_len;
			}

			for (i = 1, carry = 0; i < move_len; ++i) {
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				--xp; --yp;
				carry	+= (digit2int(xp) + digit2int(yp));
				*rp	= 	int2digit(carry % 10);
				carry	/= 10;
			}

			/* set decimal point */
			if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			--xp; --yp; *rp	= *xp;
		}
	}
	else {
		/* `x' and `y' has no fraction part */
		xp = x->ep; yp = y->ep; carry = 0;
	}

	/* process integer parts */
	x_len = (size_t)(xp - x->ip); y_len = (size_t)(yp - y->ip);
	move_len	= ((x_len < y_len)?	x_len: y_len);
	for (i = 0; i < move_len; ++i) {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		--xp; --yp;
		carry	+= (digit2int(xp) + digit2int(yp));
		*rp		= int2digit(carry % 10);
		carry	/= 10;
	}
	if (y_len > x_len) {
		/* 'y's integer part is longer than `x's */
		xp	= yp; /* process `y's integer part */
		move_len	= y_len - move_len;
	}
	else {
		/* 'x's integer part is longer than `y's */
		move_len	= x_len - move_len;
	}
	for (i = 0; i < move_len; ++i) {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		--xp; --yp;
		carry	+= digit2int(xp);
		*rp		= int2digit(carry % 10);
		carry	/= 10;
	}
	if (carry != 0) {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		*rp		= int2digit(carry);
		carry	= 0;
	}

	/* set zero ('0'), if integer part is empty */
	if ((*rp == '\0') || *rp == '.') {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		*rp	= int2digit(0);
	}

	/* move the result string to the left */
	if (rp > result) {	strcpy(result, rp);	}

	return	result; /* return the result buffer address */
}

/* subtract two Number_t structures `x' and `y'
 * signs of `x' and `y' must be Sign_e_positive
 * `x' should be greater than `y', or `x' should equals `y'
 */
/*@null@*/ static char *sub(Number_t	*x
							, Number_t	*y
							, char		*result
							, size_t	result_size
							)
{
	const char	*xp, *yp;
	size_t		x_len, y_len, len, i;
	char		*rp;
	int			bollow;

	rp = result + (result_size - 1); *rp = '\0';

	/* process the fraction part */
	if ((x->dp != x->ep) || y->dp != y->ep) {
		/* `x' or `y' has fraction parts */
		x_len = (size_t)(x->ep - x->dp); y_len = (size_t)(y->ep - y->dp);
		if (x_len > y_len) {
			/* `x's fraction part is longer than `y's */
			len	= x_len - y_len; xp = x->ep - len; x_len -= len; yp = y->ep;
			if ((rp -= len) < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			memcpy(rp, xp, len);
			bollow = 0;
		}
		else if (x_len < y_len) {
			/* `y's fraction part is longer than `x's */
			len	= y_len - x_len; xp = x->ep; yp = y->ep; y_len -= len;
			if (x_len == 0) {
				/* we don't process the decimal point digit on `y' */
				--len;
			}
			for (i = 0, bollow = 10; i < len; bollow = 9, ++i) {
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				*rp	= int2digit(bollow - digit2int(--yp));
			}
			if (x_len == 0) {
				/* set the decimal point */
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				*rp	= *(--yp);
			}
			bollow	= 1;
		}
		else {
			/* `x' is such long as `y's */
			xp = x->ep; yp = y->ep; bollow = 0;
		}

		/* subtract fraction parts of `x' and 'y' */
		if (x_len > 0) {
			len = x_len - 1;
			for (i = 0; i < len; ++i) {
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				if ((bollow = digit2int(--xp) - bollow - digit2int(--yp))
																	< 0) {
					*rp	= int2digit(10 + bollow);	bollow = 1;
				}
				else {
					*rp	= int2digit(bollow);		bollow = 0;
				}
			}
			if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			*rp	= *(--xp); --yp; /* set the decimal point */
		}
	}
	else {
		/* `x' and `y' has no fraction parts */
		xp = x->dp; yp = y->dp; bollow = 0;
	}

	/* process the integer parts */
	x_len = (size_t)(x->dp - x->ip); y_len = (size_t)(y->dp - y->ip);
	for (i = 0; i < y_len; ++i) {
		if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
		}
		if ((bollow = digit2int(--xp) - bollow - digit2int(--yp)) < 0) {
			*rp	= int2digit(10 + bollow);	bollow = 1;
		}
		else {
			*rp	= int2digit(bollow);		bollow = 0;
		}
	}
	len	= x_len - y_len;
	for (i = 0; i < len; ++i) {
		if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
		}
		if ((bollow = digit2int(--xp) - bollow) < 0) {
			*rp	= int2digit(10 + bollow);	bollow = 1;
		}
		else {
			*rp	= int2digit(bollow);		bollow = 0;
		}
	}

	/* truncate zero codes */
	while (((*rp != '\0') && *(rp + 1) != '.') && *rp == '0') {	++rp;	}

	/* set zero ('0'), if integer part is empty */
	if ((*rp == '\0') || *rp == '.') {
		if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
		}
		*rp	= int2digit(0);
	}

	/* move the result string to the left */
	if (rp > result) {	strcpy(result, rp);	}

	return	result;	/* return the result buffer address */
}

/* multiplies two Number_t structures `x' and `y'
 * signs of `x' and `y' must be Sign_e_positive
 */
/*@null@*/ static char *mul(Number_t	*x
							, Number_t	*y
							, char		*result
							, size_t	result_size
							)
{
	static int	muls[10][10] =	{ {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 }
								, {  0,  1,  2,  3,  4,  5,  6,  7,  8,  9 }
								, {  0,  2,  4,  6,  8, 10, 12, 14, 16, 18 }
								, {  0,  3,  6,  9, 12, 15, 18, 21, 24, 27 }
								, {  0,  4,  8, 12, 16, 20, 24, 28, 32, 36 }
								, {  0,  5, 10, 15, 20, 25, 30, 35, 40, 45 }
								, {  0,  6, 12, 18, 24, 30, 36, 42, 48, 54 }
								, {  0,  7, 14, 21, 28, 35, 42, 49, 56, 63 }
								, {  0,  8, 16, 24, 32, 40, 48, 56, 64, 72 }
								, {  0,  9, 18, 27, 36, 45, 54, 63, 72, 81 }
								};
	struct segment_t {
		const char	*bp, *ep; /* begin and end pointer */
	}			xseg[2], yseg[2];
	const char	*xp, *yp;
	size_t		i, j;
	char		*rp, *sp, *rbp;
	int			carry, yval;

	/* initialize `result' buffer */
	rbp = rp = result + (result_size - 1); *rp = '\0';
	if ((result_size - 1) > 0) {
		memset(result, (int)'0', (result_size - 1));
	}

	/* initialize integer/fraction segments of `x' and `y' */
	xseg[0].bp = x->dp;		xseg[0].ep = x->ep - 1;
	xseg[1].bp = x->ip - 1;	xseg[1].ep = x->dp - 1;
	yseg[0].bp = y->dp;		yseg[0].ep = y->ep - 1;
	yseg[1].bp = y->ip - 1;	yseg[1].ep = y->dp - 1;

	/* process the numbers */
	for (i = 0; i < (sizeof(yseg) / sizeof(yseg[0])); ++i) {
		for (yp = yseg[i].ep; yp > yseg[i].bp; --yp, --rp) {
			sp = rp; yval = digit2int(yp); carry = 0;
			for (j = 0; j < (sizeof(xseg) / sizeof(xseg[0])); ++j) {
				for (xp = xseg[j].ep; xp > xseg[j].bp; --xp) {
					if (--sp < result) {
						/* Error: insufficient size of result */
						errno = ERANGE; return NULL;
					}

					/* multiplies the digits */
					carry += (muls[yval][digit2int(xp)] + digit2int(sp));
					*sp = int2digit(carry % 10); carry /= 10;
				}
			}

			/* process carry */
			while (carry > 0) {
				if (--sp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				*sp = int2digit(carry % 10); carry /= 10;
			}

			/* set most left digit of result */
			if (rbp > sp) {	rbp	= sp;	}
		}
	}

	/* construct the result string */
	/* get end pointer to result buffer (`rp') */
	rp	= result + (result_size - 1);
	/* get fraction length of result buffer (`i') */
	if ((i = (size_t)(x->ep - x->dp)) != 0) {	--i;	}
	if ((j = (size_t)(y->ep - y->dp)) != 0) {	i += (--j);	}
	/* get pointer to the fraction part (`sp') */
	sp	= rp - i;
	if (rbp < sp) {
		/* result has integer part */
		j	= (size_t)(sp - rbp); /* get integer part length */
		if (rbp != result) {
			memcpy(result, rbp, j); /* move integer part */
		}
		if (i != 0) {
			/* result has fraction part */
			if ((result + (j + 1)) > sp) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}

			result[j]	= '.'; /* put decimal point */
			if ((result + (j + 1)) < sp) {
				/* move fraction part */
				memcpy((result + (j + 1)), sp, (i + 1));
			}
		}
		else {
			/* result has no fraction part */
			result[j]	= '\0'; /* set end of string */
		}
	}
	else {
		/* result has no integer part */
		if ((result + 2) > sp) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}

		/* set leading string "0." to result buffer */
		memcpy(result, "0.", 2);
		/* move the fraction part (or end of string) */
		if ((result + 2) != sp) {
			memcpy((result + 2), sp, i + 1);
		}
	}

	return	result; /* returns result buffer */
}

/* allocate the result buffer (sums[k]) */
/*@null@*/ static Result_t *allocate_result_buffer(Result_t		*result
													, Number_t	*a
													, Number_t	*b
													)
{
	size_t	size = ((((a->dp - a->ip) > (b->dp - b->ip))?
									(a->dp - a->ip): (b->dp - b->ip))
						+ (((a->ep - a->dp) > (b->ep - b->dp))?
									(a->ep - a->dp): (b->ep - b->dp))
						+ (size_t)4
						);
	char	*bufp;

	if (result->size < size) {
		if ((bufp = malloc(size)) == NULL) {
			/* Error: insufficient memory */
			return	NULL;
		}

		if (result->bufp != NULL) {
			memset(result->bufp, (int)'\0', result->size);
			free(result->bufp);
		}
		result->size = size; result->bufp = bufp;
	}

	return	result;
}

/* API: aggregate the numeral strings
 * this returns address of the allocated result buffer
 */
/*@null@*/ char *aggregate_numbers(Aggregate_e_opcode	opcode
									, const char		**numerals
									, size_t			n_numerals
									, ssize_t			scale
									)
{
	Result_t	sums[2];
	Number_t	a, b;
	size_t		i, j, k, result_left_size;
	char		*retval, *resultp;
	int			sv_errno = errno;

	/* test: arguments are invalid */
	if ((((numerals == NULL)
	|| n_numerals < 1)
	|| opcode < Aggregate_e_opcode_add)
	|| opcode >= Aggregate_e_opcode_n
	) {
		errno = EINVAL; return NULL; /* Error: invalid arguments */
	}

	/* initialize the result buffer */
	sums[1].size = (size_t)0; sums[1].bufp = NULL;

	/* set the first numeral to `result' buffer */
	if ((sums[0].bufp = malloc((sums[0].size = (strlen(numerals[0]) + 1))))
																== NULL) {
		return	NULL; /* Error: insufficient memories */
	}
	retval	= strcpy(sums[0].bufp, numerals[0]);

	/* aggregate numerals[1] to numerals[n_numerals]... */
	for (i = 1, j = 0; (retval != NULL) && i < n_numerals; ++i, j = k) {
		/* test: arguments are invalid? */
		if (((numerals[i] == NULL)
		|| make_number(sums[j].bufp,	&a) == NULL)
		|| make_number(numerals[i],		&b) == NULL
		) {
			/* Error: invalid arguments */
			errno = EINVAL;
CLEAN_UP:	/* release the allocated buffers of sums[] */
			if (sums[0].bufp != NULL) {
				memset(sums[0].bufp, (int)'\0', sums[0].size);
				free(sums[0].bufp);
				sums[0].bufp = NULL;sums[0].size = (size_t)0;
			}
			if (sums[1].bufp != NULL) {
				memset(sums[1].bufp, (int)'\0', sums[1].size);
				free(sums[1].bufp);
				sums[1].bufp = NULL;sums[1].size = (size_t)0;
			}

			retval = NULL; break; /* Error: returns NULL */
		}

		/* set index of `sums[]' as result of calculating */
		k	= ((j + 1) % (sizeof(sums) / sizeof(sums[0])));

		if (opcode == Aggregate_e_opcode_mul) {
			/* allocate the result buffer (sums[k]) */
			result_left_size = (
							(((a.dp - a.ip) == 0)? 1: (a.dp - a.ip))
							+ (((b.dp - b.ip) == 0)? 1: (b.dp - b.ip))
							+ (((a.ep - a.dp) == 0)? 0: ((a.ep - a.dp) - 1))
							+ (((b.ep - b.dp) == 0)? 0: ((b.ep - b.dp) - 1))
							+ (size_t)3
							);
			if (sums[k].size < result_left_size) {
				if ((resultp = malloc(result_left_size)) == NULL) {
					/* Error: can't allocate the result buffer */
					errno = ENOMEM; goto CLEAN_UP;
				}
				memset(sums[k].bufp, (int)'\0', sums[k].size);
				free(sums[k].bufp);
				sums[k].size = result_left_size; sums[k].bufp = resultp;
			}

			if ((is_zero(&a) != 0) || is_zero(&b) != 0) {
				/* if `a' or `b' is zero, result is zero */
				if (sums[k].size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; goto CLEAN_UP;
				}
				retval	= strcpy(sums[k].bufp, "0");
			}
			else {
				if (a.sign != b.sign) {
					if (sums[k].size == 1) {
						/* Error: insufficient size of `result' */
						errno = ERANGE; goto CLEAN_UP;
					}
					resultp = sums[k].bufp; *resultp = '-'; ++resultp
					; result_left_size = sums[k].size - 1
					;
				}
				else {
					resultp = sums[k].bufp; result_left_size = sums[k].size;
				}

				retval	= mul(&a, &b, resultp, result_left_size);
			}
		}
		else if (((opcode == Aggregate_e_opcode_sub) && a.sign != b.sign)
		|| ((opcode == Aggregate_e_opcode_add) && a.sign == b.sign)
		) {
			Sign_e	sign;

			/* allocate the result buffer (sums[k]) */
			if (allocate_result_buffer((sums + k), &a, &b) == NULL) {
				/* Error: insufficient memory */
				errno = ENOMEM; goto CLEAN_UP;
			}

			sign = a.sign; a.sign = b.sign = Sign_e_positive;
			if (sign == Sign_e_negative) {
				if (sums[k].size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; goto CLEAN_UP;
				}

				sums[k].bufp[0]			= '-'
				; resultp			= (sums[k].bufp + 1)
				; result_left_size = sums[k].size - 1
				;
			}
			else {
				resultp				= sums[k].bufp
				; result_left_size	= sums[k].size
				; 
			}
			retval	= add(&a, &b, resultp, result_left_size);
		}
		else if ((((opcode == Aggregate_e_opcode_sub)
					&& a.sign == Sign_e_positive)
					&& b.sign == Sign_e_positive)
		|| (((opcode == Aggregate_e_opcode_add)
					&& a.sign == Sign_e_positive)
					&& b.sign == Sign_e_negative)
		) {
			/* allocate the result buffer (sums[k]) */
			if (allocate_result_buffer((sums + k), &a, &b) == NULL) {
				/* Error: insufficient memory */
				errno = ENOMEM; goto CLEAN_UP;
			}

			a.sign = b.sign = Sign_e_positive;
			if (compare(&a, &b) < 0) {
				if (sums[k].size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; goto CLEAN_UP;
				}

				sums[k].bufp[0]	= '-';
				retval	= sub(&b, &a
								, (sums[k].bufp + 1), (sums[k].size - 1));
			}
			else {
				retval	= sub(&a, &b, sums[k].bufp, sums[k].size);
			}
		}
		else {
			/* allocate the result buffer (sums[k]) */
			if (allocate_result_buffer((sums + k), &a, &b) == NULL) {
				/* Error: insufficient memory */
				errno = ENOMEM; goto CLEAN_UP;
			}

			a.sign = b.sign = Sign_e_positive;
			if (compare(&b, &a) < 0) {
				if (sums[k].size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; goto CLEAN_UP;
				}

				sums[k].bufp[0]	= '-';
				retval	= sub(&a, &b
							, (sums[k].bufp + 1), (sums[k].size - 1)
							);
			}
			else {
				retval	= sub(&b, &a, sums[k].bufp, sums[k].size);
			}
		}
	}

	/* if allocated buffer contains the last result,
	 * we must copy it to `result' buffer
	 */
	if ((retval != NULL) && j != 0) {
		Result_t	tmp_sum = sums[0];
		sums[0] = sums[1]; sums[1] = tmp_sum;
	}

	/* release the allocated buffer */
	memset(sums[1].bufp, (int)'\0', sums[1].size)
	; free((void *)(sums[1].bufp))
	; sums[1].bufp = NULL; sums[1].size = (size_t)0
	;

	/* set the result numeral string, according to scale */
	if ((retval != NULL) && scale >= (size_t)0) {
		if ((resultp = strchr(sums[0].bufp, '.')) == NULL) {
			/* the result has no fraction part */
			result_left_size	= strlen(sums[0].bufp);
			if ((resultp
					= malloc((result_left_size + 1 + (size_t)scale + 1)))
																== NULL) {
				/* Error: insufficient memory */
INSUFFICIENT_RESULT_MEMORY:
				errno = ENOMEM;
				memset(sums[0].bufp, (int)'\0', sums[0].size);
				free(sums[0].bufp);
				sums[0].bufp = NULL; sums[0].size = (size_t)0;

				return	NULL; /* Error:  return NULL */
			}
			strcpy(resultp, sums[0].bufp);
			if (scale > 0) {
				resultp[result_left_size]	= '.';
				memset((resultp + (result_left_size + 1))
						, (int)'0', (size_t)scale
						);
			}
		}
		else {
			/* the result has the fraction part */
			result_left_size	= (size_t)(resultp - sums[0].bufp);
			if ((resultp = malloc(result_left_size + 1 + scale + 1))
																== NULL) {
				/* Error: insufficient memory */
				goto INSUFFICIENT_RESULT_MEMORY;
			}
			memcpy(resultp, sums[0].bufp, result_left_size);
			if ((i = strlen(sums[0].bufp + result_left_size))
													> (size_t)scale + 1) {
				memcpy(resultp + result_left_size
						, sums[0].bufp + result_left_size
						, (size_t)(scale + 1)
						);
			}
			else {
				strcpy(resultp + result_left_size
						, sums[0].bufp + result_left_size
						);
				if (i != ((size_t)scale + 1)) {
					memset(resultp + (result_left_size + i)
							, (int)'0'
							, (size_t)scale - i + 1
							);
				}
			}
		}
		resultp[result_left_size + 1 + scale]	= '\0';
		/* release the buffer pointed from sums[0].bufp */
		memset(sums[0].bufp, (int)'\0', sums[0].size);
		free(sums[0].bufp);
		sums[0].bufp	= resultp; /* set the allocated result buffer */
	}

	/* if several errors happened, this API returns NULL.
	 * otherwise, this returns address of `result' buffer
	 */
	return	((retval != NULL)? ((errno = sv_errno), sums[0].bufp): NULL);
}

/* API: release the buffer that is allocated by `aggregate_numbers()' */
int destroy_aggregate_numbers(char	*aggregate_numbers_result)
{
	int	sv_errno = errno;

	if (aggregate_numbers_result == NULL) {
		errno = EINVAL; return !0; /* Error: invalid arguments */
	}

	memset(aggregate_numbers_result
			, (int)'\0'
			, strlen(aggregate_numbers_result)
			);
	free(aggregate_numbers_result);

	errno = sv_errno; return 0; /* normal end */
}

/* API: compare two numeral strings */
int *compare_numbers(const char	*a_str, const char	*b_str, int	*result)
{
	Number_t	a, b;
	int			sv_errno = errno;

	/* test: arguments are invalid
	 * convert the numeral strings to Number_t structures
	 */
	if (((result == NULL)
	|| make_number(a_str, &a) == NULL)
	|| make_number(b_str, &b) == NULL
	) {
		errno = EINVAL; return NULL; /* Error: invalid arguments */
	}

	/* return result of compare */
	*result	= compare(&a, &b); errno = sv_errno; return result;
}
